#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  configfile.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import yaml
from wishbone.external.attrdict import AttrDict
from jsonschema import validate

SCHEMA = {
    "type": "object",
    "properties": {
        "lookups": {
            "type": "object",
            "patternProperties": {
                ".*": {
                    "type": "object",
                    "properties": {
                        "module": {
                            "type": "string"
                        },
                        "arguments": {
                            "type": "object"
                        }
                    },
                    "required": ["module"],
                    "additionalProperties": False
                }
            }
        },
        "modules": {
            "type": "object",
            "patternProperties": {
                ".*": {
                    "type": "object",
                    "properties": {
                        "module": {
                            "type": "string"
                        },
                        "description": {
                            "type": "string"
                        },
                        "arguments": {
                            "type": "object"
                        }
                    },
                    "required": ["module"],
                    "additionalProperties": False
                }
            }
        },
        "routingtable": {
            "type": "array"
        }
    },
    "required": ["modules", "routingtable"],
    "additionalProperties": False
}


class ConfigFile(object):

    def __init__(self, filename, logstyle):
        self.logstyle = logstyle
        self.config = {"lookups": {}, "modules": {}, "routingtable": []}
        self.__addLogFunnel()
        self.__addMetricFunnel()
        self.load(filename)

    def addModule(self, name, module, arguments={}, description="", context="configfile"):

        if name.startswith('_'):
            raise Exception("Module instance names cannot start with _.")

        if name not in self.config["modules"]:
            self.config["modules"][name] = {'description': description, 'module': module, 'arguments': arguments, 'context': context}
            self.addConnection(name, "logs", "_logs", name, context="_logs")
            self.addConnection(name, "metrics", "_metrics", name, context="_metrics")

        else:
            raise Exception("Module instance name '%s' is already taken." % (name))

    def addLookup(self, name, module, arguments={}):

        if name not in self.config["lookups"]:
            self.config["lookups"][name] = {"module": module, "arguments": arguments}
        else:
            raise Exception("Uplook instance name '%s' is already taken." % (name))

    def addConnection(self, source_module, source_queue, destination_module, destination_queue, context="configfile"):

        connected = self.__queueConnected(source_module, source_queue)

        if not connected:
            self.config["routingtable"].append(AttrDict({"source_module": source_module, "source_queue": source_queue, "destination_module": destination_module, "destination_queue": destination_queue, "context": context}))
        else:
            raise Exception("Cannot connect '%s.%s' to '%s.%s'. Reason: %s." % (source_module, source_queue, destination_module, destination_queue, connected))

    def dump(self):
        return AttrDict(self.config)

    def load(self, filename):

        config = self.__load(filename)
        self.__validate(config)
        self.__validateRoutingTable(config)

        if "lookups" in config:
            for lookup in config["lookups"]:
                self.addLookup(name=lookup, **config["lookups"][lookup])

        for module in config["modules"]:
            self.addModule(name=module, **config["modules"][module])

        for route in config["routingtable"]:
            sm, sq, dm, dq = self.__splitRoute(route)
            self.addConnection(sm, sq, dm, dq)

        getattr(self, "_setupLogging%s" % (self.logstyle.upper()))()

    def __queueConnected(self, module, queue):

        for c in self.config["routingtable"]:
            if (c["source_module"] == module and c["source_queue"] == queue) or (c["destination_module"] == module and c["destination_queue"] == queue):
                return "Queue '%s.%s' is already connected to '%s.%s'" % (c["source_module"], c["source_queue"], c["destination_module"], c["destination_queue"])
        return False

    def __splitRoute(self, route):

        (source, destination) = route.split('->')
        (source_module, source_queue) = source.rstrip().lstrip().split('.')
        (destination_module, destination_queue) = destination.rstrip().lstrip().split('.')
        return source_module, source_queue, destination_module, destination_queue

    def __load(self, filename):
        '''Loads and returns the yaml bootstrap file.'''

        try:
            with open(filename, 'r') as f:
                config = yaml.load(f)
        except Exception as err:
            raise Exception("Failed to load bootstrap file.  Reason: %s" % (err))
        else:
            return config

    def __validate(self, config):

        try:
            validate(config, SCHEMA)
        except Exception as err:
            raise Exception("Failed to validate configuration file.  Reason: %s" % (err.message))

    def __validateRoutingTable(self, config):

        for route in config["routingtable"]:
            (left, right) = route.split("->")
            assert "." in left.lstrip().rstrip(), "routingtable rule \"%s\" does not have the right format. Missing a dot." % (route)
            assert "." in right.lstrip().rstrip(), "routingtable rule \"%s\" does not have the right format. Missing a dot." % (route)

    def __addLogFunnel(self):

        self.config["modules"]["_logs"] = {'description': "Centralizes the logs of all modules.", 'module': "wishbone.flow.funnel", "arguments": {}, "context": "_logs"}

    def __addMetricFunnel(self):

        self.config["modules"]["_metrics"] = {'description': "Centralizes the metrics of all modules.", 'module': "wishbone.flow.funnel", "arguments": {}, "context": "_metrics"}

    def _setupLoggingSTDOUT(self):

        if not self.__queueConnected("_logs", "outbox"):
            self.config["modules"]["_logs_format"] = {'description': "Create a human readable log format.", 'module': "wishbone.encode.humanlogformat", "arguments": {}, "context": "_logs"}
            self.addConnection("_logs", "outbox", "_logs_format", "inbox", context="_logs")
            self.config["modules"]["_logs_stdout"] = {'description': "Prints all incoming logs to STDOUT.", 'module': "wishbone.output.stdout", "arguments": {}, "context": "_logs"}
            self.addConnection("_logs_format", "outbox", "_logs_stdout", "inbox", context="_logs")

    def _setupLoggingSYSLOG(self):

        if not self.__queueConnected("_logs", "outbox"):
            self.config["modules"]["_logs_syslog"] = {'description': "Writes all incoming messags to syslog.", 'module': "wishbone.output.syslog", "arguments": {}, "context": "_logs"}
            self.addConnection("_logs", "outbox", "_logs_syslog", "inbox", context="_logs")




#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
# template.py
#
#  Copyright 2013 Jelle Smet <jelle.smet@tomtom.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from jinja2 import Environment, FileSystemLoader
from jinja2 import Template as JinjaTemplate
from wishbone import Actor


class Template(Actor):

    '''**A Wishbone module which generates a text from a dictionary and a template.**

    Converts a dictionary to a text using the defined Jinja2 template.

    Optionally header template values can be converted too.


    Parameters:

        - location(str)("./")
           |  The directory containing templates.

        - template(str)()*
           |  The template filename stored in directory <location>.

        - source(str)("@data")*
           |  The dictionary to use.

        - destination(str)("@data")*
           |  The location to which the rendered result has to be stored.

        - header_templates(dict)({})*
           |  A dict of templates to render. Can be lookup values.
           |  Results are stored in this module's instance header
           |  using the provided key.



    Queues:

        - inbox
           |  Incoming events

        - outbox
           |  Outgoing events

    '''

    def __init__(self, actor_config, location="./", template=None, source="@data", destination="@data", header_templates={}):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        if self.kwargs.template is not None:
            self.templates = Environment(loader=FileSystemLoader(self.kwargs.location))

    def consume(self, event):

        event = self.construct(event)
        self.submit(event, self.pool.queue.outbox)

    def construct(self, event):

        for name, template in self.kwargs.header_templates:
            try:
                template_r = JinjaTemplate(template)
                result = template_r.render(**event.get())
                event.set(result, "@tmp.%s.%s" % (self.name, name))
            except Exception as err:
                self.logging.warning(
                    "Failed to convert header key '%s'.  Reason: %s" % (name, err))
                raise

        if self.kwargs.template is not None:
            try:
                template = self.templates.get_template(self.kwargs.template)
            except Exception as err:
                self.logging.error('No template found with filename "%s%s".' % (self.kwargs.location, self.kwargs.template))
                raise
            else:
                try:
                    result = template.render(**event.get(self.kwargs.source))
                    event.set(result, self.kwargs.destination)
                except Exception as err:
                    self.logging.error('There was an error processing the template. Reason: %s' % (err))
                    raise
        return event



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  loglevelfilter.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Log


class LogLevelFilter(Actor):

    '''**Filters log events based on loglevel.**

    Drops internal Wishbone logs based on loglevel.

    Internal Wishbone format:

    (6, 1367682301.430527, 'Router', 'Received SIGINT. Shutting down.')


    Parameters:

        - loglevel(int)(6)
           |  The maximum allowed loglevel.


    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, loglevel=6):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consumer, "inbox")

    def consumer(self, event):

        data = event.get()
        if isinstance(data, Log):
            if data.level <= self.kwargs.loglevel:
                self.submit(event, self.pool.queue.outbox)
        else:
            self.logging.warning("Incoming event is not of type <wishbone.event.Log>. Dropped")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  jsonvalidate.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from json import loads
from jsonschema import validate


class JSONValidate(Actor):

    '''**Validates JSON data against JSON-schema.**

    Validates the a Python dictionary (converted from a JSON string) against a
    predefined JSONschema. http://json-schema.org/

    The defined schema has to be valid JSON data.

    Events which do not pass validation are send to the default <failed> queue.

    Parameters:

        - schema(str)(None)
           |  The filename of the JSON validation schema to load.  When no
           |  schema is defined no validation is done.


    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, schema=None):

        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):
        if self.kwargs.schema is not None:
            self.logging.debug("Validation schema defined.  Doing validation.")
            self.schema_data = self.__loadValidationSchema(self.kwargs.schema)
            self.validate = self.__validate
        else:
            self.logging.debug("No validation schema defined.  No validation.")
            self.validate = self.__noValidate

    def consume(self, event):

        data = event.get()

        try:
            self.validate(data)
        except Exception as err:
            raise Exception(self.__condenseError(err))
        else:
            self.submit(event, self.pool.queue.outbox)

    def __loadValidationSchema(self, path):
        with open(path, 'r') as schema:
            data = ''.join(schema.readlines())
            return loads(data)

    def __validate(self, data):

        return validate(data, self.schema_data)

    def __noValidate(self, data):
        return True

    def __condenseError(self, err):

        return " -> ".join([l.lstrip() for l in str(err).split("\n") if l != ""])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  influxdb.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Metric


class InfluxDB(Actor):

    '''**Converts the internal metric format to InfluxDB line format.**

    Incoming date must be of type <Metric>.


    Parameters:

        - script(bool)(True)
           |  Include the script name.

        - pid(bool)(False)
           |  Include pid value in script name.

        - source(bool)(True):
           |  Include the source name in the naming schema.


    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")
        self.tags = {}

    def consume(self, event):

        data = event.get()

        if isinstance(data, Metric):
            event.set(self.generate(data))
            self.submit(event, self.pool.queue.outbox)
        else:
            self.logging.error("Event dropped because not of type <wishbone.event.Metric>")

    def generate(self, data):

        name = data.name.split('.')
        return "{name},source={source},type={type},module={module},queue={queue} value={value:.5f} {time:20.0f}".format(
                    name=name[4],
                    source=data.source,
                    type=data.type,
                    module=name[1],
                    queue=name[3],
                    value=data.value,
                    time=data.time * 1000000000)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  event.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import arrow
import time

EVENT_RESERVED = ["@timestamp", "@version", "@data", "@tmp", "@errors"]


class Log(object):

    '''
    A Wishbone log object
    '''

    def __init__(self, time, level, pid, module, message):

        self.time = time
        self.level = level
        self.pid = pid
        self.module = module
        self.message = message

    def __str__(self):

        return "Log(%s)" % (self.__dict__)

    def __repr__(self):

        return "Log(%s)" % (self.__dict__)


class Metric(object):

    '''
    A Wishbone metric object
    '''

    def __init__(self, time, type, source, name, value, unit, tags):

        self.time = time
        self.type = type
        self.source = source
        self.name = name
        self.value = value
        self.unit = unit
        self.tags = tags

    def __str__(self):

        return "Metric(%s)" % (self.__dict__)

    def __repr__(self):

        return "Metric(%s)" % (self.__dict__)

    def dump(self):

        return self.__dict__


class Event(object):

    '''
    **The Wishbone event object representation.**

    A class object containing the event data being passed from one Wishbone
    module to the other.
    '''

    def __init__(self, data=None):

        self.data = {
            "@timestamp": time.time(),
            "@version": 1,
            "@data": data,
            "@tmp": {
            },
            "@errors": {
            }
        }

    def clone(self):
        '''
        Returns a cloned version of the event using deepcopy.
        '''

        c = self.deepish_copy(self.data)
        e = Event()
        e.data = c
        return e

    def copy(self, source, destination):
        '''
        Copies the source key to the destination key.

        :param str source: The name of the source key.
        :param str destination: The name of the destination key.
        '''

        self.set(self.deepish_copy(self.get(source)), destination)

    def delete(self, key=None):
        '''
        Deletes a key.

        :param str key: The name of the key to delete
        '''

        s = key.split('.')
        if s[0] in EVENT_RESERVED and len(s) == 1:
            raise Exception("Cannot delete root of reserved keyword '%s'." % (key))

        if key is None:
            self.data = None
        else:
            if '.' in key:
                s = key.split('.')
                key = '.'.join(s[:-1])
                del(self.get(key)[s[-1]])
            else:
                del(self.data[key])

    def get(self, key="@data"):
        '''
        Returns the value of <key>.

        :param str key: The name of the key to read.
        :return: The value of <key>
        '''

        def travel(path, d):

            if len(path) == 1:
                if isinstance(d, dict):
                    return d[path[0]]
                else:
                    raise Exception()
            else:
                return travel(path[1:], d[path[0]])
        if key is None or key is "":
            return self.data
        else:
            path = key.split('.')
            try:
                return travel(path, self.data)
            except:
                raise KeyError(key)

    def set(self, value, key="@data"):
        '''
        Sets the value of <key>.

        :param value: The value to set.
        :param str key: The name of the key to assign <value> to.
        '''

        if key.startswith('@') and key.split('.')[0] not in EVENT_RESERVED:
            raise Exception("Keys starting with @ are reserved.")
        result = value
        for name in reversed(key.split('.')):
            result = {name: result}

        self.dict_merge(self.data, result)
        # self.data.update(result)

    def dump(self, complete=False, convert_timestamp=True):
        '''
        Dumps the content of the event.

        :param bool complete: Determines whether to include @tmp and @errors.
        :param bool convert_timestamp: When True converts <Arrow> object to iso8601 string.
        :return: The content of the event.
        :rtype: dict
        '''

        d = {}
        for key, value in self.data.iteritems():
            if key == "@tmp" and not complete:
                continue
            if key == "@errors" and not complete:
                continue
            elif isinstance(value, arrow.arrow.Arrow) and convert_timestamp:
                d[key] = str(value)
            else:
                d[key] = value

        return d

    raw = dump

    def dict_merge(self, dct, merge_dct):

        """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
        updating only top-level keys, dict_merge recurses down into dicts nested
        to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
        ``dct``.

        Stolen from https://gist.github.com/angstwad/bf22d1822c38a92ec0a9

        :param dct: dict onto which the merge is executed
        :param merge_dct: dct merged into dct
        :return: None
        """
        for k, v in merge_dct.iteritems():
            if k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], dict):
                self.dict_merge(dct[k], merge_dct[k])
            else:
                dct[k] = merge_dct[k]

    def deepish_copy(self, org):
        '''
        much, much faster than deepcopy, for a dict of the simple python types.

        Blatantly ripped off from https://writeonly.wordpress.com/2009/05/07
        /deepcopy-is-a-pig-for-simple-data/
        '''

        if isinstance(org, dict):
            out = dict().fromkeys(org)
            for k, v in org.iteritems():
                try:
                    out[k] = v.copy()   # dicts, sets
                except AttributeError:
                    try:
                        out[k] = v[:]   # lists, tuples, strings, unicode
                    except TypeError:
                        out[k] = v      # ints

            return out
        else:
            return org



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  fanout.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor


class Fanout(Actor):

    '''**Forward each incoming message to all connected queues.**

    Forward each incoming message to all connected queues.

    Parameters:

        n/a


    Queues:

        inbox
         |  Outgoing events.

    '''

    def __init__(self, actor_config):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):
        self.destinations = []
        for queue in self.pool.listQueues(names=True, default=False):
            if queue != "inbox":
                self.destinations.append(self.pool.getQueue(queue))

    def consume(self, event):

        for queue in self.destinations:
            self.submit(event.clone(), queue)




"""
AttrDict is a mapping object that allows access to its values both as
keys, and as attributes (whenever the key can be used as an attribute
name).
"""
from collections import Mapping, MutableMapping, Sequence
import json
import re
from sys import version_info


__all__ = ['AttrDict', 'merge']


# Python 2
PY2, STRING = (True, basestring) if version_info < (3,) else (False, str)


class AttrDict(MutableMapping):
    """
    A mapping object that allows access to its values both as keys, and
    as attributes (as long as the attribute name is valid).
    """
    def __init__(self, mapping=None, recursive=True,
                 default_factory=None, pass_key=False):
        """
        mapping: (optional, None) The mapping object to use for the
            instance. Note that the mapping object itself is used, not a
            copy. This means that you cannot clone an AttrDict using:
            adict = AttrDict(adict)
        recursive: (optional, True) Recursively convert mappings into
            AttrDicts.
        default_factory: (optional, Not Passed) If passed make AttrDict
            behave like a default dict.
        pass_key: (optional, False) If True, and default_factory is
            given, then default_factory will be passed the key.
        """
        if mapping is None:
            mapping = {}

        if default_factory is None:
            self.__setattr__('_default_factory', None, force=True)
            self.__setattr__('_pass_key', False, force=True)
        else:
            self.__setattr__('_default_factory', default_factory, force=True)
            self.__setattr__('_pass_key', pass_key, force=True)

        self.__setattr__('_recursive', recursive, force=True)

        self.__setattr__('_mapping', mapping, force=True)

        for key, value in mapping.iteritems() if PY2 else mapping.items():
            if self._valid_name(key):
                setattr(self, key, value)

    def get(self, key, default=None):
        """
        Get a value associated with a key.

        key: The key associated with the desired value.
        default: (optional, None) The value to return if the key is not
            found.
        """
        return self._mapping.get(key, default)

    def items(self):
        """
        In python 2.X returns a list of (key, value) pairs as 2-tuples.
        In python 3.X returns an iterator over the (key, value) pairs.
        """
        return self._mapping.items()

    def keys(self):
        """
        In python 2.X returns a list of keys in the mapping.
        In python 3.X returns an iterator over the mapping's keys.
        """
        return self._mapping.keys()

    def values(self):
        """
        In python 2.X returns a list of values in the mapping.
        In python 3.X returns an iterator over the mapping's values.
        """
        return self._mapping.values()

    def _set(self, key, value):
        """
        Responsible for actually adding/changing a key-value pair. This
        needs to be separated out so that setattr and setitem don't
        clash.
        """
        self._mapping[key] = value

        if self._valid_name(key):
            super(AttrDict, self).__setattr__(
                key, self._build(value, recursive=self._recursive))

    def _delete(self, key):
        """
        Responsible for actually deleting a key-value pair. This needs
        to be separated out so that delattr and delitem don't clash.
        """
        del self._mapping[key]

        if self._valid_name(key):
            super(AttrDict, self).__delattr__(key)

    def __getattr__(self, key):
        """
        value = adict.key

        Access a value associated with a key in the instance.
        """
        if self._default_factory is None or key.startswith('_'):
            raise AttributeError(key)

        return self.__missing__(key)

    def __call__(self, key):
        """
        Access a value in the mapping as an attribute. This differs from
        dict-style key access because it returns a new instance of an
        AttrDict (if the value is a mapping object), not the underlying
        type. This allows for dynamic attribute-style access.

        key: The key associated with the value being accessed.
        """
        if key not in self._mapping:
            if self._default_factory is not None:
                self.__missing__(key)
            else:
                raise AttributeError(
                    "'{0}' instance has no attribute '{1}'".format(
                        self.__class__.__name__, key))

        return self._build(self._mapping[key], recursive=self._recursive)

    def __setattr__(self, key, value, force=False):
        """
        adict.key = value

        Add a key-value pair as an attribute
        """
        if force:
            super(AttrDict, self).__setattr__(key, value)
        else:
            if not self._valid_name(key):
                raise TypeError("Invalid key: {0}".format(repr(key)))

            self._set(key, value)

    def __delattr__(self, key):
        """
        del adict.key

        Remove a key-value pair as an attribute.
        """
        if not self._valid_name(key) or key not in self._mapping:
            raise TypeError("Invalid key: {0}".format(repr(key)))

        self._delete(key)

    def __setitem__(self, key, value):
        """
        adict[key] = value

        Add a key-value pair to the instance.
        """
        self._set(key, value)

    def __getitem__(self, key):
        """
        value = adict[key]

        Access a value associated with a key in the instance.
        """
        if key not in self._mapping and self._default_factory is not None:
            self[key] = self.__missing__(key)

        return self._mapping[key]

    def __delitem__(self, key):
        """
        del adict[key]

        Delete a key-value pair in the instance.
        """
        self._delete(key)

    def __contains__(self, key):
        """
        key in adict

        Check if a key is in the instance.
        """
        return key in self._mapping

    def __len__(self):
        """
        len(adict)

        Check the length of the instance.
        """
        return len(self._mapping)

    def __iter__(self):
        """
        (key for key in adict)

        iterate through all the keys.
        """
        return self._mapping.__iter__()

    def __add__(self, other):
        """
        adict + other

        Add a mapping to this AttrDict object.

        NOTE: AttrDict is not commutative. a + b != b + a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        recursive = not hasattr(other, '_recursive') or other._recursive

        return merge(self, other, recursive=self._recursive and recursive)

    def __radd__(self, other):
        """
        other + adict

        Add this AttrDict to a mapping object.

        NOTE: AttrDict is not commutative. a + b != b + a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        return merge(other, self, recursive=self._recursive)

    def __repr__(self):
        """
        Create a string representation of the AttrDict.
        """
        return u"AttrDict({0})".format(repr(self._mapping))

    def __missing__(self, key):
        """
        Add a missing element.
        """
        if self._pass_key:
            self._mapping[key] = value = self._default_factory(key)
        else:
            self._mapping[key] = value = self._default_factory()

        return value

    def __getstate__(self):
        """
        Handle proper serialization of the object. (used by pickle).
        """
        return (self._mapping, self._recursive, self._default_factory,
                self._pass_key)

    def __setstate__(self, state):
        """
        Handle proper deserialization of the object. (used by pickle).
        """
        mapping, recursive, default_factory, pass_key = state
        self.__init__(mapping, recursive, default_factory, pass_key)

    @classmethod
    def _build(cls, obj, recursive=True):
        """
        Wrap an object in an AttrDict as necessary. Mappings are
        wrapped, but all other objects are returned as is.

        obj: The object to (possibly) wrap.
        recursive: (optional, True) Whether Sequences should have their
            elements turned into attrdicts.
        """
        if isinstance(obj, Mapping):
            obj = cls(obj, recursive=recursive)
        elif recursive:
            if isinstance(obj, Sequence) and not isinstance(obj, STRING):
                new = [cls._build(element, recursive=True) for element in obj]

                # This has to have a __class__. Only old-style classes
                # don't, and none of them would subclass Sequence.
                obj = obj.__class__(new)

        return obj

    @classmethod
    def _valid_name(cls, name):
        """
        Check whether a key name is a valid attribute name. A valid
        name must start with an alphabetic character, and must only
        contain alphanumeric characters and underscores. The name also
        must not be an attribute of this class.

        NOTE: Names with leading underscores are considered invalid for
        stylistic reasons. While this package is fairly un-Pythonic, I'm
        going to stand strong on the fact that leading underscores
        represent private attributes. Further, magic methods absolutely
        need to be prevented so that crazy things don't happen.
        """
        return (isinstance(name, STRING) and
                re.match('^[A-Za-z][A-Za-z0-9_]*$', name) and
                not hasattr(cls, name))

    # Add missing iter methods in 2.X
    if PY2:
        def iteritems(self):
            """
            Iterate over (key, value) 2-tuples in the mapping
            """
            for key, value in self._mapping.iteritems():
                if isinstance(value, dict):
                    yield key, self._build(value, recursive=self._recursive)
                else:
                    yield key, value

        def iterkeys(self):
            """
            Iterate over keys in the mapping
            """
            return self._mapping.iterkeys()

        def itervalues(self):
            """
            Iterate over values in the mapping
            """
            for value in self._mapping.itervalues():
                if isinstance(value, dict):
                    yield self._build(value, recursive=self._recursive)
                else:
                    yield value

        def has_key(self, key):
            """
            True if the mapping has key, False otherwise
            """
            return key in self


def merge(left, right, recursive=True):
    """
    merge to mappings objects into a new AttrDict.

    left: The left mapping object.
    right: The right mapping object.
    recursive: (optional, True) Whether Sequences should have their
        elements turned into attrdicts.

    NOTE: This is not commutative. merge(a, b) != merge(b, a).
    """
    merged = AttrDict(recursive=recursive)

    left_keys = set(left)
    right_keys = set(right)

    # Items only in the left object
    for key in (left_keys - right_keys):
        merged[key] = left[key]

    # Items only in the right object
    for key in (right_keys - left_keys):
        merged[key] = right[key]

    # In both
    for key in left_keys.intersection(right_keys):
        if isinstance(left[key], Mapping) and isinstance(right[key], Mapping):
            merged[key] = merge(left[key], right[key], recursive=recursive)
        else:  # different types, overwrite with the right value
            merged[key] = right[key]

    return merged


# def load(*filenames, load_function=json.load)  # once Python3-only
def load(*filenames, **kwargs):
    """
    Returns a settings dict built from a list of settings files.

    filenames: The names of any number of settings files.
    load_function: (optional, json.load) The function used to load the
    settings into a Mapping object.
    """
    load_function = kwargs.pop('load_function', json.load)

    if kwargs:
        raise TypeError("unknown options: {0}".format(kwargs.keys()))

    settings = AttrDict()

    for filename in filenames:
        with open(filename, 'r') as fileobj:
            settings += load_function(fileobj)

    return settings



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  gearmanin.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from gevent import monkey; monkey.patch_all()
from wishbone import Actor
from wishbone.event import Event
from gevent import sleep
from gearman import GearmanWorker
from Crypto.Cipher import AES
import base64


class GearmanIn(Actor):

    '''**Consumes events/jobs from  Gearmand.**

    Consumes jobs from a Gearmand server.
    When secret is none, no decryption is done.


    Parameters:

        - hostlist(list)(["localhost:4730"])
           |  A list of gearmand servers.  Each entry should have
           |  format host:port.

        - secret(str)(None)
           |  The AES encryption key to decrypt Mod_gearman messages.

        - workers(int)(1)
           |  The number of gearman workers within 1 process.

        - queue(str)(wishbone)
           |  The queue to consume jobs from.


    Queues:

        - outbox:   Outgoing events.

    '''

    def __init__(self, actor_config, hostlist=["localhost:4730"], secret=None, workers=1, queue="wishbone"):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("outbox")
        self.background_instances = []

        if self.kwargs.secret is None:
            self.decrypt = self.__plainTextJob
        else:
            key = self.kwargs.secret[0:32]
            self.cipher = AES.new(key + chr(0) * (32 - len(key)))
            self.decrypt = self.__encryptedJob

    def preHook(self):
        for _ in range(self.kwargs.workers):
            self.sendToBackground(self.gearmanWorker)

    def consume(self, gearman_worker, gearman_job):

        decrypted = self.decrypt(gearman_job.data)
        event = Event(decrypted)
        self.submit(event, self.pool.queue.outbox)
        return decrypted

    def __encryptedJob(self, data):
        return self.cipher.decrypt(base64.b64decode(data))

    def __plainTextJob(self, data):
        return data

    def gearmanWorker(self):

        self.logging.info("Gearmand worker instance started")
        while self.loop():
            try:
                worker_instance = GearmanWorker(self.kwargs.hostlist)
                worker_instance.register_task(self.kwargs.queue, self.consume)
                worker_instance.work()
            except Exception as err:
                self.logging.warn('Connection to gearmand failed. Reason: %s. Retry in 1 second.' % err)
                sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  zmqtopicout.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone import Actor
import zmq.green as zmq


class ZMQTopicOut(Actor):

    '''**Publishes data to one or more ZeroMQ Topic subscribe modules.**

    Expects wishbone.input.topic modules to take the initiative and connect to
    this module.  The clients subscribe to a topic of interest.  When multiple
    subscribers subscriber to the same topic, they will all receive the same
    messages resulting into a fanout messaging pattern.

    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event

        - port(int)(19283)
           |  The port to bind to.

        - timeout(int)(1)
           |  The time in seconds to timeout when connecting.

        - topic(str)("")*
           |  The default topic to use when none is set in the header.


    Queues:

        - inbox
           |  Incoming events submitted to the outside.

    '''

    def __init__(self, actor_config, selection='@data', port=19283, timeout=10, topic=""):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PUB)
        self.socket.bind("tcp://*:%s" % self.kwargs.port)

    def consume(self, event):

        data = event.get(self.kwargs.selection)
        try:
            self.socket.send("%s %s" % (self.kwargs.topic, data))
        except Exception as err:
            raise ("Failed to submit message.  Reason %s" % (err))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  tcpout.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from gevent import socket


class TCPOut(Actor):

    '''**A TCP client which writes data to a TCP socket.**

    Writes data to a tcp socket.

    When <data> is of type list, all elements
    will be joined using <delimiter> and submitted together.

    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - host(string)("localhost")
           |  The host to submit to.

        - port(int)(19283)
           |  The port to submit to.

        - timeout(int)(1)
           |  The time in seconds to timeout when connecting

        - delimiter(str)("\\n")*
           |  A delimiter to add to each event.


    Queues:

        - inbox
           |  Incoming events submitted to the outside.

    '''

    def __init__(self, actor_config, selection="@data", host="127.0.0.1", port=19283, timeout=10, delimiter="\n"):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def consume(self, event):

        data = event.get(self.kwargs.selection)
        if isinstance(data, list):
            data = self.kwargs.delimiter.join(data)

        connection = self.setupConnection()
        connection.sendall(str(data) + self.kwargs.delimiter)
        connection.close()

    def setupConnection(self):

        s = socket.socket()
        s.settimeout(self.kwargs.timeout)
        s.connect((self.kwargs.host, self.kwargs.port))
        return s



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  namedpipein.py
#
#  Copyright 2016 Jelle Smet development@smetj.net
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Event
from gevent import sleep
from gevent import os as gevent_os

# from gevent import monkey;monkey.patch_all()
import os


class NamedPipeIn(Actor):

    '''**Takes data in from a named pipe..**

    Creates a named pipe and read data from it.

    Parameters:

        - path(str)("/tmp/wishbone")
           |  The the location of the named pipe.

    Queues:

        - outbox
           |  Data coming from the outside world.
    '''

    def __init__(self, actor_config, path="/tmp/wishbone"):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("outbox")

    def preHook(self):

        os.mkfifo(self.kwargs.path)
        self.logging.info('Named pipe %s created.' % (self.kwargs.path))
        self.sendToBackground(self.drain, self.kwargs.path)

    def consume(self, event):
        for line in event:
            e = Event(line)
            self.submit(e, self.pool.queue.outbox)

    def drain(self, p):
        '''Reads the named pipe.'''

        self.logging.info('Started.')
        fd = os.open(p, os.O_RDWR | os.O_NONBLOCK)
        gevent_os.make_nonblocking(fd)

        while self.loop():
            try:
                lines = gevent_os.nb_read(fd, 4096).splitlines()
                if len(lines) == 0:
                    sleep(0.5)
                else:
                    self.consume(lines)
            except OSError:
                pass

    def postHook(self):

        try:
            os.unlink(self.kwargs.path)
        except:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  ircin.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from gevent import monkey; monkey.patch_all()
from wishbone import Actor

import irc.bot
import irc.strings
from irc.client import ip_numstr_to_quad, ip_quad_to_numstr


class IRCBot(Actor):

    '''**Joins an IRC channel to accept input.**

    A simple IRC bot which joins a channel, reads all public and private
    messages and sends these to the module's queues.


    Parameters:

        - server(str)("localhost")
           |  The IRC server to connect to.

        - port(int)(6667)
           |  The IRC server port to connect to.

        - nickname(str)("wishbone")
           |  The bot's nickname.

        - channels(list)(["wishbone"])
           |  The list of channels to join.
           |  Each channel is mapped to a queue.
           |  The queue name has the # portion stripped off.

    Queues:

        - outbox
           |  The messages of all channels.

        - <channels>

           |  For each channel in <channels> a queue is created receiving the
           |  messages of only that channel.
    '''

    def __init__(self, actor_config, server="localhost", port=6667, nickname="wishbone", channels=["wishbone"]):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("outbox")

        for channel in self.kwargs.channels:
            self.pool.createQueue(channel)

    def preHook(self):





class TestBot(irc.bot.SingleServerIRCBot):
    def __init__(self, channels, nickname, server, port=6667):
        irc.bot.SingleServerIRCBot.__init__(self, [(server, port)], nickname, nickname)
        self.channel = channel

    def on_nicknameinuse(self, c, e):
        c.nick(c.get_nickname() + "_")

    def on_welcome(self, c, e):
        c.join(self.channel)

    def on_privmsg(self, c, e):
        self.do_command(e, e.arguments[0])

    def on_pubmsg(self, c, e):
        a = e.arguments[0].split(":", 1)
        if len(a) > 1 and irc.strings.lower(a[0]) == irc.strings.lower(self.connection.get_nickname()):
            self.do_command(e, a[1].strip())
        return

    def on_dccmsg(self, c, e):
        # non-chat DCC messages are raw bytes; decode as text
        text = e.arguments[0].decode('utf-8')
        c.privmsg("You said: " + text)

    def on_dccchat(self, c, e):
        if len(e.arguments) != 2:
            return
        args = e.arguments[1].split()
        if len(args) == 4:
            try:
                address = ip_numstr_to_quad(args[2])
                port = int(args[3])
            except ValueError:
                return
            self.dcc_connect(address, port)

    def do_command(self, e, cmd):
        nick = e.source.nick
        c = self.connection

        if cmd == "disconnect":
            self.disconnect()
        elif cmd == "die":
            self.die()
        elif cmd == "stats":
            for chname, chobj in self.channels.items():
                c.notice(nick, "--- Channel statistics ---")
                c.notice(nick, "Channel: " + chname)
                users = sorted(chobj.users())
                c.notice(nick, "Users: " + ", ".join(users))
                opers = sorted(chobj.opers())
                c.notice(nick, "Opers: " + ", ".join(opers))
                voiced = sorted(chobj.voiced())
                c.notice(nick, "Voiced: " + ", ".join(voiced))
        elif cmd == "dcc":
            dcc = self.dcc_listen()
            c.ctcp("DCC", nick, "CHAT chat %s %d" % (
                ip_quad_to_numstr(dcc.localaddress),
                dcc.localport))
        else:
            c.notice(nick, "Not understood: " + cmd)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  default.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.actor import ActorConfig
from wishbone.error import ModuleInitFailure, NoSuchModule
from gevent import signal, event, sleep, spawn
import multiprocessing
import importlib
from gevent import pywsgi
import json
from .graphcontent import GRAPHCONTENT
from .graphcontent import VisJSData


class Container():
    pass


class ModulePool():

    def __init__(self):

        self.module = Container()

    def list(self):
        '''Returns a generator returning all module instances.'''

        for m in self.module.__dict__.keys():
            yield self.module.__dict__[m]

    def getModule(self, name):
        '''Returns a module instance'''

        try:
            return getattr(self.module, name)
        except AttributeError:
            raise NoSuchModule("Could not find module %s" % name)


class Default(multiprocessing.Process):

    '''The default Wishbone router.

    Holds all Wishbone modules and connections.

    Arguments:

        - router_config(obj)            : The router setup configuration.

        - module_manager(obj)           : A Wishbone ModuleManager object instance.

        - size(int)(100)                : The size of all queues.

        - frequency(int)(1)             : The frequency at which metrics are produced.

        - identification                : A string identifying this instance in logging.

        - stdout_logging(bool)(True)    : When True all logs are written to STDOUT.

        - background(bool)(False)       : When True, sends the router to background in a
                                          separate process.


    '''

    def __init__(self, router_config, module_manager, size=100, frequency=1, identification="wishbone", stdout_logging=True, process=False, graph=False, graph_include_sys=False):

        if process:
            multiprocessing.Process.__init__(self)
            self.daemon = True
        self.config = router_config
        self.module_manager = module_manager
        self.size = size
        self.frequency = frequency
        self.identification = identification
        self.stdout_logging = stdout_logging
        self.process = process
        self.graph = graph
        self.graph_include_sys = graph_include_sys

        self.module_pool = ModulePool()

        self.__running = False
        self.__block = event.Event()
        self.__block.clear()

        signal(2, self.initiateStop)
        signal(15, self.__noop)

        self.initiate_stop = event.Event()
        self.initiate_stop.clear()
        spawn(self.stop)

    def block(self):
        '''Blocks until stop() is called.'''
        self.__block.wait()

    def getChildren(self, module):
        children = []

        def lookupChildren(module, children):
            for module in self.module_pool.getModule(module).getChildren():
                name = module.split(".")[0]
                if name not in children:
                    children.append(name)
                    lookupChildren(name, children)

        lookupChildren(module, children)
        return children

    def isRunning(self):
        return self.__running

    def start(self):
        '''Starts all registered modules.'''

        if self.process:
            self.run = self.__start
            multiprocessing.Process.start(self)
            return self
        else:
            self.__start()

    def initiateStop(self, *args, **kwargs):

        self.initiate_stop.set()

    def stop(self):
        '''Stops all modules.'''

        self.initiate_stop.wait()
        for module in self.module_pool.list():
            if module.name not in self.getChildren("_logs") + ["_logs"] and not module.stopped:
                module.stop()

        while not self.__logsEmpty():
            sleep(0.1)

        # This gives an error when starting in background, no idea why
        # self.module_pool.module.wishbone_logs.stop()

        self.__running = False
        self.__block.set()

    def __connect(self, source, destination):
        '''Connects one queue to the other.

        For convenience, the syntax of the queues is <modulename>.<queuename>
        For example:

            stdout.inbox
        '''

        (source_module, source_queue) = source.split('.')
        (destination_module, destination_queue) = destination.split('.')

        source = self.module_pool.getModule(source_module)
        destination = self.module_pool.getModule(destination_module)

        source.connect(source_queue, destination, destination_queue)

    def __initConfig(self):
        '''Setup all modules and routes.'''

        lookup_modules = {}

        for name, instance in self.config.lookups.iteritems():
            lookup_modules[name] = self.__registerLookupModule(instance.module, **instance.arguments)

        for name, instance in self.config.modules.iteritems():
            pmodule = self.module_manager.getModuleByName(instance.module)

            if instance.description == "":
                instance.description = pmodule.__doc__.split("\n")[0].replace('*', '')

            actor_config = ActorConfig(name, self.size, self.frequency, lookup_modules, instance.description)

            self.__registerModule(pmodule, actor_config, instance.arguments)

        self.__setupConnections()

    def __logsEmpty(self):
        '''Checks each module whether any logs have stayed behind.'''

        for module in self.module_pool.list():
            if not module.pool.queue.logs.size() == 0:
                return False
        else:
            return True

    def __noop(self, *args, **kwargs):
        pass

    def __registerLookupModule(self, module, **kwargs):

        base = ".".join(module.split('.')[0:-1])
        function = module.split('.')[-1]
        m = importlib.import_module(base)
        return getattr(m, function)(**kwargs)

    def __registerModule(self, module, actor_config, arguments={}):
        '''Initializes the wishbone module module.'''

        try:
            setattr(self.module_pool.module, actor_config.name, module(actor_config, **arguments))
        except Exception as err:
            raise ModuleInitFailure("Problem loading module %s.  Reason: %s" % (actor_config.name, err))

    def __setupConnections(self):
        '''Setup all connections as defined by configuration_manager'''

        for route in self.config.routingtable:
            self.__connect("%s.%s" % (route.source_module, route.source_queue), "%s.%s" % (route.destination_module, route.destination_queue))

    def __start(self):

        self.__initConfig()
        self.__running = True

        if self.graph:
            self.graph = GraphWebserver(self.config, self.module_pool, self.__block, self.graph_include_sys)
            self.graph.start()

        for module in self.module_pool.list():
            module.start()

        self.block()


class GraphWebserver():

    def __init__(self, config, module_pool, block, include_sys):
        self.config = config
        self.module_pool = module_pool
        self.block = block
        self.js_data = VisJSData()

        for c in self.config["routingtable"]:
            if self.__include(include_sys, self.config["modules"][c.source_module]["context"], self.config["modules"][c.destination_module]["context"]):
                self.js_data.addModule(instance_name=c.source_module,
                                       module_name=self.config["modules"][c.source_module]["module"],
                                       description=self.module_pool.getModule(c.source_module).description)

                self.js_data.addModule(instance_name=c.destination_module,
                                       module_name=self.config["modules"][c.destination_module]["module"],
                                       description=self.module_pool.getModule(c.destination_module).description)

                self.js_data.addQueue(c.source_module, c.source_queue)
                self.js_data.addQueue(c.destination_module, c.destination_queue)
                self.js_data.addEdge("%s.%s" % (c.source_module, c.source_queue), "%s.%s" % (c.destination_module, c.destination_queue))

    def __include(self, include_sys, source_module_context, destination_module_context):

        if include_sys:
            return True
        elif source_module_context not in ["_logs", "_metrics"] and destination_module_context not in ["_logs", "_metrics"]:
            return True
        else:
            return False

    def start(self):

        print "#####################################################"
        print "#                                                   #"
        print "# Caution: Started webserver on port 8088           #"
        print "#                                                   #"
        print "#####################################################"
        spawn(self.setupWebserver)

    def stop(self):
        pass

    def loop(self):

        return not self.block.isSet()

    def getMetrics(self):

        def getConnectedModuleQueue(m, q):
            for c in self.config["routingtable"]:
                if c.source_module == m and c.source_queue == q:
                    return (c.destination_module, c.destination_queue)
            return (None, None)

        d = {"module": {}}
        for module in self.module_pool.list():
            d["module"][module.name] = {}
            for queue in module.pool.listQueues(names=True):
                d["module"][module.name]["queue"] = {queue: {"metrics": module.pool.getQueue(queue).stats()}}
                (dest_mod, dest_q) = getConnectedModuleQueue(module.name, queue)
                if dest_mod is not None and dest_q is not None:
                    d["module"][module.name]["queue"] = {queue: {"connection": {"module": dest_mod, "queue": dest_q}}}
        return json.dumps(d)

    def application(self, env, start_response):
        if env['PATH_INFO'] == '/':
            start_response('200 OK', [('Content-Type', 'text/html')])
            return[GRAPHCONTENT % (self.js_data.dumpString()[0], self.js_data.dumpString()[1])]
        elif env['PATH_INFO'] == '/metrics':
            start_response('200 OK', [('Content-Type', 'text/html')])
            return[self.getMetrics()]
        else:
            start_response('404 Not Found', [('Content-Type', 'text/html')])
            return [b'<h1>Not Found</h1>']

    def setupWebserver(self):

        pywsgi.WSGIServer(('', 8088), self.application, log=None, error_log=None).serve_forever()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  bootstrap.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#



import argparse
import os
# http://stackoverflow.com/questions/4554271/how-to-avoid-excessive-stat-etc-localtime-calls-in-strftime-on-linux
os.environ["TZ"] = ":/etc/localtime"
import sys

from wishbone.router import Default
from wishbone import ModuleManager
from wishbone.config import ConfigFile
from wishbone.utils import PIDFile
from gevent import signal
from daemon import DaemonContext
from pkg_resources import get_distribution
from jinja2 import Template


class BootStrap():

    '''
    Parses command line arguments and bootstraps the Wishbone instance.
    '''

    def __init__(self, description="Wishbone bootstrap server. Build composable event pipeline servers with minimal effort.", include_groups=[]):

        parser = argparse.ArgumentParser(description=description)
        subparsers = parser.add_subparsers(dest='command')

        start = subparsers.add_parser('start', description="Starts a Wishbone instance and detaches to the background.  Logs are written to syslog.")
        start.add_argument('--config', type=str, dest='config', default='wishbone.cfg', help='The Wishbone bootstrap file to load.')
        start.add_argument('--instances', type=int, dest='instances', default=1, help='The number of parallel Wishbone instances to bootstrap.')
        start.add_argument('--pid', type=str, dest='pid', default='%s/wishbone.pid' % (os.getcwd()), help='The pidfile to use.')
        start.add_argument('--queue-size', type=int, dest='queue_size', default=100, help='The queue size to use.')
        start.add_argument('--frequency', type=int, dest='frequency', default=1, help='The metric frequency.')
        start.add_argument('--id', type=str, dest='identification', default=None, help='An identification string.')
        start.add_argument('--module_path', type=str, dest='module_path', default=None, help='A comma separated list of directories to search and find Wishbone modules.')

        debug = subparsers.add_parser('debug', description="Starts a Wishbone instance in foreground and writes logs to STDOUT.")
        debug.add_argument('--config', type=str, dest='config', default='wishbone.cfg', help='The Wishbone bootstrap file to load.')
        debug.add_argument('--instances', type=int, dest='instances', default=1, help='The number of parallel Wishbone instances to bootstrap.')
        debug.add_argument('--queue-size', type=int, dest='queue_size', default=100, help='The queue size to use.')
        debug.add_argument('--frequency', type=int, dest='frequency', default=1, help='The metric frequency.')
        debug.add_argument('--id', type=str, dest='identification', default=None, help='An identification string.')
        debug.add_argument('--module_path', type=str, dest='module_path', default=None, help='A comma separated list of directories to search and find Wishbone modules.')
        debug.add_argument('--graph', action="store_true", help='When enabled starts a webserver on 8088 showing a graph of connected modules and queues.')
        debug.add_argument('--graph_include_sys', action="store_true", help='When enabled includes logs and metrics related queues modules and queues to graph layout.')

        debug.add_argument('--profile', action="store_true", help='When enabled profiles the process and dumps a profile file in the current directory. The profile file can be loaded in Chrome developer tools.')

        stop = subparsers.add_parser('stop', description="Tries to gracefully stop the Wishbone instance.")
        stop.add_argument('--pid', type=str, dest='pid', default='wishbone.pid', help='The pidfile to use.')

        kill = subparsers.add_parser('kill', description="Kills the Wishbone processes immediately.")
        kill.add_argument('--pid', type=str, dest='pid', default='wishbone.pid', help='The pidfile to use.')

        llist = subparsers.add_parser('list', description="Lists the available Wishbone modules.")
        llist.add_argument('--group', type=str, dest='group', default=None, help='List the modules of this group type.')

        show = subparsers.add_parser('show', description="Shows the details of a module.")
        show.add_argument('--module', type=str, required=True, help='Shows the documentation of the module. ')

        arguments = vars(parser.parse_args())

        if arguments["command"] == "list":
            if arguments["group"] is not None:
                arguments["include_groups"] = [arguments["group"]]
            else:
                arguments["include_groups"] = include_groups

        dispatch = Dispatch()
        getattr(dispatch, arguments["command"])(**arguments)


class Dispatch():

    '''
    Handles the Wishbone instance commands.
    '''

    def __init__(self):

        self.routers = []
        self.__stopping = False

    def generateHeader(self):
        '''Generates the Wishbone ascii header.'''

        with open("%s/data/banner.tmpl" % (os.path.dirname(__file__))) as f:
            template = Template(''.join(f.readlines()))

        return template.render(version=get_distribution('wishbone').version)

    def debug(self, command, config, instances, queue_size, frequency, identification, module_path, graph, graph_include_sys, profile):
        '''
        Handles the Wishbone debug command.
        '''

        if module_path is not None:
            self.__expandSearchPath(module_path)

        processes = []

        def stopSequence(*args, **kwargs):
            for proc in processes:
                proc.stop()

        signal(2, stopSequence)

        module_manager = ModuleManager()
        router_config = ConfigFile(config, 'STDOUT').dump()

        if instances == 1:
            sys.stdout.write("\nInstance started in foreground with pid %s\n" % (os.getpid()))

            if profile:
                from wishbone.utils.py2devtools import Profiler
                with Profiler():
                    Default(router_config, module_manager, size=queue_size, frequency=frequency, identification=identification, stdout_logging=True, graph=graph, graph_include_sys=graph_include_sys).start()
            else:
                Default(router_config, module_manager, size=queue_size, frequency=frequency, identification=identification, stdout_logging=True, graph=graph, graph_include_sys=graph_include_sys).start()

        else:
            for instance in range(instances):
                processes.append(Default(router_config, module_manager, size=queue_size, frequency=frequency, identification=identification, stdout_logging=True, process=True).start())
            pids = [str(p.pid) for p in processes]
            print(("\nInstances started in foreground with pid %s\n" % (", ".join(pids))))
            for proc in processes:
                proc.join()

    def list(self, command, group, category=None, include_groups=[]):

        print((self.generateHeader()))
        print("Available modules:")
        print((ModuleManager().getModuleTable(category, group, include_groups)))

    def show(self, command, module):
        '''
        Shows the help message of a module.
        '''

        module_manager = ModuleManager()
        module_manager.validateModuleName(module)
        module_manager.exists(module)

        print((self.generateHeader()))
        try:
            (category, group, module) = module.split('.')
        except ValueError:
            (category, sub, group, module) = module.split('.')
            category = "%s.%s" % (category, sub)

        try:
            title = module_manager.getModuleTitle(category, group, module)
            version = module_manager.getModuleVersion(category, group, module)
            header = "%s.%s.%s" % (category, group, module)
            print("")
            print("="*len(header))
            print(header)
            print("="*len(header))
            print("")
            print("Version: %s" % (version))
            print("")
            print(title)
            print("-"*len(title))
            print(module_manager.getModuleDoc(category, group, module))
        except Exception as err:
            print("Failed to load module %s.%s.%s. Reason: %s" % (category, group, module, err))

    def start(self, command, config, instances, pid, queue_size, frequency, identification, module_path):
        '''
        Handles the Wishbone start command.
        '''

        if module_path is not None:
            self.__expandSearchPath(module_path)

        module_manager = ModuleManager()
        router_config = ConfigFile(config, 'SYSLOG').dump()
        pid_file = PIDFile(pid)

        with DaemonContext(stdout=sys.stdout, stderr=sys.stderr, files_preserve=self.__getCurrentFD(), detach_process=True):
            if instances == 1:
                sys.stdout.write("\nWishbone instance started with pid %s\n" % (os.getpid()))
                sys.stdout.flush()
                pid_file.create([os.getpid()])
                Default(router_config, module_manager, size=queue_size, frequency=frequency, identification=identification, stdout_logging=False).start()
            else:
                processes = []
                for instance in range(instances):
                    processes.append(Default(router_config, module_manager, size=queue_size, frequency=frequency, identification=identification, stdout_logging=False, process=True).start())
                pids = [str(p.pid) for p in processes]
                print(("\n%s Wishbone instances started in background with pid %s\n" % (len(pids), ", ".join(pids))))
                pid_file.create(pids)
                for proc in processes:
                    proc.join()

    def stop(self, command, pid):
        '''
        Handles the Wishbone stop command.
        '''

        try:
            pid = PIDFile(pid)
            sys.stdout.write("Stopping instance with PID ")
            sys.stdout.flush()
            for entry in pid.read():
                sys.stdout.write(" %s " % (entry))
                sys.stdout.flush()
                pid.sendSigint(entry)
            pid.cleanup()
            print("")
        except Exception as err:
            print("")
            print(("Failed to stop instances.  Reason: %s" % (err)))

    def __stopSequence(self):
        '''
        Calls the stop() function of each instance.
        '''

        if not self.__stopping:
            # TODO: Weird hack, otherwise when trapping signal(2) this function is
            #      executed many times.
            self.__stopping = True
            for instance in self.routers:
                if hasattr(instance, "stop"):
                    instance.stop()

    def __getCurrentFD(self):
        '''
        returns a list with filedescriptors in use.
        '''

        try:
            return [int(x) for x in os.listdir("/proc/self/fd")]
        except Exception as err:
            print(("Failed to get active filedescriptors.  Reason: %s." % (err)))
            sys.exit(1)

    def __alive(self, pid):
        try:
            os.kill(pid, 0)
            return True
        except:
            False

    def __expandSearchPath(self, module_path):
        for d in module_path.split(','):
            sys.path.append(d.strip())


def main():
    try:
        BootStrap()
    except Exception as err:
        print("Failed to bootstrap instance.  Reason: %s" % (err))

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  jsondecode.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from json import loads


class JSONDecode(Actor):

    '''**Decodes JSON data to Python data objects.**

    Decodes the payload or complete events from JSON format.

    Parameters:

        - source(str)("@data")
           |  The source of the event to decode.
           |  Use an empty string to refer to the complete event.

        - destination(str)("@data")
           |  The destination key to store the Python <dict>.
           |  Use an empty string to refer to the complete event.

        - unicode(bool)(True)
           |  When True, converts strings to unicode otherwise regular string.


    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, source="@data", destination="@data", unicode=True):

        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        if self.kwargs.unicode:
            self.convert = self.doUnicode
        else:
            self.convert = self.doNoUnicode

    def consume(self, event):

        data = event.get(self.kwargs.source)
        data = self.convert(data)
        event.set(data, self.kwargs.destination)
        self.submit(event, self.pool.queue.outbox)

    def doUnicode(self, data):
        return loads(data)

    def doNoUnicode(self, data):
        return self.json_loads_byteified(data)

    def json_loads_byteified(self, json_text):
        return self._byteify(
            loads(json_text, object_hook=self._byteify),
            ignore_dicts=True
        )

    def _byteify(self, data, ignore_dicts=False):
        # if this is a unicode string, return its string representation
        if isinstance(data, unicode):
            return data.encode('utf-8')
        # if this is a list of values, return list of byteified values
        if isinstance(data, list):
            return [ self._byteify(item, ignore_dicts=True) for item in data ]
        # if this is a dictionary, return dictionary of byteified keys and values
        # but only if we haven't already byteified it
        if isinstance(data, dict) and not ignore_dicts:
            return {
                self._byteify(key, ignore_dicts=True): self._byteify(value, ignore_dicts=True)
                for key, value in data.iteritems()
            }
        # if it's anything else, return it in its original form
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  setup.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.queue import QueuePool
from wishbone.logging import Logging
from wishbone.event import Event as Wishbone_Event
from wishbone.event import Metric
from wishbone.error import QueueConnected, ModuleInitFailure
from wishbone.lookup import EventLookup
from uplook.errors import NoSuchValue
from collections import namedtuple
from gevent import spawn, kill
from gevent import sleep, socket
from gevent.event import Event
from wishbone.error import QueueFull
from time import time
from sys import exc_info
from uplook import UpLook
import traceback
import inspect

Greenlets = namedtuple('Greenlets', "consumer generic log metric")
ActorConfig = namedtuple('ActorConfig', 'name size frequency lookup description')


class Actor():

    def __init__(self, config):

        self.config = config
        self.name = config.name
        self.size = config.size
        self.frequency = config.frequency
        self.description = config.description

        self.pool = QueuePool(config.size)

        self.logging = Logging(config.name, self.pool.queue.logs)

        self.__loop = True
        self.greenlets = Greenlets([], [], [], [])
        self.greenlets.metric.append(spawn(self.metricProducer))

        self.__run = Event()
        self.__run.clear()

        self.__connections = {}

        self.__children = {}
        self.__parents = {}

        self.__lookups = {}

        self.__buildUplook()

        self.stopped = True

    def connect(self, source, destination_module, destination_queue):
        '''Connects the <source> queue to the <destination> queue.
        In fact, the source queue overwrites the destination queue.'''

        if source in self.__children:
            raise QueueConnected("Queue %s.%s is already connected to %s." % (self.name, source, self.__children[source]))
        else:
            self.__children[source] = "%s.%s" % (destination_module.name, destination_queue)

        if destination_queue in destination_module.__parents:
            raise QueueConnected("Queue %s.%s is already connected to %s" % (destination_module.name, destination_queue, destination_module.__parents[destination_queue]))
        else:
            destination_module.__parents[destination_queue] = "%s.%s" % (self.name, source)

        if not self.pool.hasQueue(source):
            self.pool.createQueue(source)

        setattr(destination_module.pool.queue, destination_queue, self.pool.getQueue(source))
        self.pool.getQueue(source).disableFallThrough()
        self.logging.debug("Connected queue %s.%s to %s.%s" % (self.name, source, destination_module.name, destination_queue))

    def createEvent(self):

        '''
        Convenience function which returns an empty Wishbone event with the
        current namespace already set.
        '''

        return Wishbone_Event()

    def doEventLookup(self, name):
        (n, t, k) = name.split('.')

        try:
            return self.current_event.get(name)
        except AttributeError:
            return "You should use a dynamic lookup ~~ for header lookups. "
        except KeyError:
            self.logging.debug("There is no lookup value with name '%s'." % (name))
            raise NoSuchValue

    def getChildren(self, queue=None):
        '''Returns the queue name <queue> is connected to.'''

        if queue is None:
            return [self.__children[q] for q in self.__children.keys()]
        else:
            return self.__children[queue]

    def loop(self):
        '''The global lock for this module'''

        return self.__loop

    def postHook(self):

        pass

    def preHook(self):

        self.logging.debug("Initialized.")

    def registerConsumer(self, function, queue):
        '''Registers <function> to process all events in <queue>

        Do not trap errors.  When <function> fails then the event will be
        submitted to the "failed" queue,  If <function> succeeds to the
        success queue.'''

        self.greenlets.consumer.append(spawn(self.__consumer, function, queue))

    def start(self):
        '''Starts the module.'''

        if hasattr(self, "preHook"):
            self.logging.debug("preHook() found, executing")
            self.preHook()

        self.__run.set()
        self.logging.debug("Started with max queue size of %s events and metrics interval of %s seconds." % (self.size, self.frequency))
        self.stopped = False

    def sendToBackground(self, function, *args, **kwargs):
        '''Executes a function and sends it to the background.'''

        self.greenlets.generic.append(spawn(function, *args, **kwargs))

    def stop(self):
        '''Stops the loop lock and waits until all registered consumers have exit otherwise kills them.'''

        self.logging.info("Received stop. Initiating shutdown.")

        self.__loop = False

        for background_job in self.greenlets.metric:
            kill(background_job)

        for background_job in self.greenlets.generic:
            kill(background_job)

        for background_job in self.greenlets.consumer:
            kill(background_job)

        if hasattr(self, "postHook"):
            self.logging.debug("postHook() found, executing")
            self.postHook()

        self.logging.debug("Exit.")

        self.stopped = True

    def submit(self, event, queue):
        '''A convenience function which submits <event> to <queue>.'''

        while self.loop():
            try:
                queue.put(event)
                break
            except QueueFull:
                sleep(0.1)

    def __consumer(self, function, queue):
        '''Greenthread which applies <function> to each element from <queue>
        '''

        self.__run.wait()

        while self.loop():
            event = self.pool.queue.__dict__[queue].get()
            self.current_event = event
            try:
                function(event)
            except Exception as err:
                exc_type, exc_value, exc_traceback = exc_info()
                info = (traceback.extract_tb(exc_traceback)[-1][1], str(exc_type), str(exc_value))
                event.set(info, "@errors.%s" % (self.name))
                self.logging.error("%s" % (err))
                self.submit(event, self.pool.queue.failed)
            else:
                self.submit(event, self.pool.queue.success)

    def __buildUplook(self):

        self.__current_event = {}
        args = {}

        for key, value in inspect.currentframe(2).f_locals.iteritems():
            if key == "self" or isinstance(value, ActorConfig):
                next
            else:
                args[key] = value

        uplook = UpLook(**args)
        for name in uplook.listFunctions():
            if name not in self.config.lookup:
                raise ModuleInitFailure("A lookup function '%s' was defined but no lookup function with that name registered." % (name))
            else:
                if isinstance(self.config.lookup[name], EventLookup):
                    uplook.registerLookup(name, self.doEventLookup)
                else:
                    uplook.registerLookup(name, self.config.lookup[name])

        self.uplook = uplook
        self.kwargs = uplook.get()

    def metricProducer(self):
        '''A greenthread which collects the queue metrics at the defined interval.'''

        self.__run.wait()
        hostname = socket.gethostname()
        while self.loop():
            for queue in self.pool.listQueues(names=True):
                for metric, value in self.pool.getQueue(queue).stats().iteritems():
                    metric = Metric(time=time(),
                                    type="wishbone",
                                    source=hostname,
                                    name="module.%s.queue.%s.%s" % (self.name, queue, metric),
                                    value=value,
                                    unit="",
                                    tags=())
                    event = Wishbone_Event(metric)
                    self.submit(event, self.pool.queue.metrics)
            sleep(self.frequency)



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
#  udsout.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from gevent import sleep, socket


class UDSOut(Actor):

    '''**Writes events to a Unix Domain Socket.**


    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - path(str)("/tmp/wishbone")
           |  The unix domain socket to write events to.

        - delimiter(str)("")
            |  The delimiter to use between each event.

    Queues:

        - inbox
           |  Incoming events.

    '''

    def __init__(self, actor_config, selection='@data', path="/tmp/wishbone", delimiter=""):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):
        self.sendToBackground(self.setupConnection)

    def consume(self, event):

        data = event.get(self.kwargs.selection)

        if isinstance(event.data, list):
            data = self.kwargs.delimiter.join(event.get())
        else:
            data = event.get()

        self.socket.send(str(data) + self.kwargs.delimiter)

    def setupConnection(self):
        while self.loop():
            try:
                self.socket.sendall('')
                sleep(1)
            except Exception as err:
                while self.loop():
                    try:
                        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                        self.socket.connect(self.kwargs.path)
                        self.logging.info("Connected to %s." % (self.kwargs.path))
                        break
                    except Exception as err:
                        self.logging.error("Failed to connect to %s. Reason: %s" % (self.kwargs.path, err))
                        sleep(1)



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
#  logging.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.event import Event, Log
from wishbone.error import QueueFull
from time import time
from os import getpid
from gevent import sleep


class MockLogger():
    '''
    A wrapper around Logging which mimics logger.Logger
    '''

    def __init__(self, name, q, level=5):

        self.level = level
        self.l = Logging(name, q)

    def flush(self):
        pass

    def write(self, line):
        self.l._Logging__log(self.level, line.rstrip())

    def writelines(self, lines):
        for line in lines:
            self.l.Logging.__log(self.level, line.rstrip())


class Logging():

    '''
    Generates Wishbone formatted log messages following the Syslog priority
    definition.
    '''

    def __init__(self, name, q):
        self.name = name
        self.logs = q

    def __log(self, level, message):

        event = Event(Log(time(), level, getpid(), self.name, message))

        while True:
            try:
                self.logs.put(event)
                break
            except QueueFull:
                sleep(0.1)

    def emergency(self, message, *args, **kwargs):
        """Generates a log message with priority emergency(0).
        """
        self.__log(0, message)
    emerg = emergency
    exception = emergency

    def alert(self, message, *args, **kwargs):
        """Generates a log message with priority alert(1).
        """
        self.__log(1, message)

    def critical(self, message, *args, **kwargs):
        """Generates a log message with priority critical(2).
        """
        self.__log(2, message)
    crit = critical

    def error(self, message, *args, **kwargs):
        """Generates a log message with priority error(3).
        """
        self.__log(3, message, *args, **kwargs)
    err = error

    def warning(self, message, *args, **kwargs):
        """Generates a log message with priority warning(4).
        """
        self.__log(4, message, *args, **kwargs)
    warn = warning

    def notice(self, message, *args, **kwargs):
        """Generates a log message with priority notice(5).
        """
        self.__log(5, message)

    def informational(self, message, *args, **kwargs):
        """Generates a log message with priority informational(6).
        """
        self.__log(6, message)
    info = informational

    def debug(self, message, *args, **kwargs):
        """Generates a log message with priority debug(7).
        """
        self.__log(7, message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       httpinserver.py
#
#       Copyright 2016 Jelle Smet development@smetj.net
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Event
from gevent import pywsgi
from gevent import socket
from gevent.pool import Pool
from wishbone.logging import MockLogger


class HTTPInServer(Actor):

    '''**Receive events over HTTP.**

    Creates an HTTP server to which events can be submitted.


    Parameters:

        - address(str)("0.0.0.0")
           |  The address to bind to.

        - port(int)(19283)
           |  The port to bind to.

        - keyfile(str)(None)
           |  When SSL is required, the location of the keyfile to use.

        - certfile(str)(None)
           |  When SSL is required, the location of the certfile to use.

        - ca_certs(str)(None)
            |  When SSL is required, the location of the ca certs to use.

        - delimiter(str)(None)
           |  The delimiter which separates multiple
           |  messages in a stream of data.

        - poolsize(int)(1000)
            |  The connection pool size.

        - so_reuseport(bool)(False)
            |  Enables socket option SO_REUSEPORT.
            |  See https://lwn.net/Articles/542629/
            |  Required when running multiple Wishbone instances.

    Queues:

        - outbox
           |  Incoming events submitted to /.


    When more queues are connected to this module instance, they are
    automatically mapped to the URL resource.

    For example http://localhost:10080/fubar is mapped to the <fubar> queue.
    The root resource "/" is mapped the <outbox> queue.
    '''

    def __init__(self, actor_config, address="0.0.0.0", port=19283, keyfile=None, certfile=None, ca_certs=None, delimiter=None, poolsize=1000, so_reuseport=False):
        Actor.__init__(self, actor_config)

        if self.kwargs.delimiter is None:
            self.delimit = self.__noDelimiter
        elif self.kwargs.delimiter == "\\n":
            self.delimit = self.__newLineDelimiter
        else:
            self.delimit = self.__otherDelimiter

        self.logger_info = MockLogger(self.name, self.pool.queue.logs, 6)
        self.logger_error = MockLogger(self.name, self.pool.queue.logs, 3)

    def preHook(self):
        self.sendToBackground(self.__serve)

    def consume(self, env, start_response):

        try:
            try:
                queue = self.__validateQueueName(env['PATH_INFO'])
            except Exception as err:
                start_response('404 Not Found', [('Content-Type', 'text/plain')])
                return "%s" % (err)

            if env["REQUEST_METHOD"] in ["PUT", "POST"]:
                counter = 0
                for line in self.delimit(env["wsgi.input"]):
                    event = Event(line)
                    self.submit(event, self.pool.getQueue(queue))
                    counter += 1
                start_response('200 OK', [('Content-Type', 'text/plain')])
                return "Submitted %s event(s) to queue '%s'." % (counter, queue)
            elif env["REQUEST_METHOD"] == "GET":
                start_response('200 OK', [('Content-Type', 'text/plain')])
                return "OK"
            else:
                start_response('405 Method Not Allowed', [('Content-Type', 'text/plain')])
                return "405 Method Not Allowed"
        except Exception as err:
            start_response('500 Internal Server Error', [('Content-Type', 'text/html')])
            return "A problem occurred processing your request. Reason: %s" % (err)

    def __validateQueueName(self, path):

        '''Validates and returns a trimmed value.'''

        result = [x for x in path.split('/') if x != ""]
        if result == []:
            queue = "outbox"
        else:
            queue = result[0]

        if len(result) > 1:
            raise Exception("Endpoint can only have a depthlevel of 1.")
        elif not self.pool.hasQueue(queue):
            raise Exception("There is no queue with name %s." % (queue))
        else:
            return queue

    def __newLineDelimiter(self, data):
        for line in data.readlines():
            yield line.rstrip()

    def __noDelimiter(self, data):
        return ["".join(data.readlines())]

    def __otherDelimiter(self, data):
        r = []
        for line in data.readlines():
            if line.rstrip("\n").endswith(self.kwargs.delimiter):
                line = line.rstrip("\n")[:-len(self.kwargs.delimiter)]
                if line != "\n":
                    r.append(line)
                yield "".join(r)
                r = []
            else:
                r.append(line)
        yield "\n".join(r)

    def __serve(self):
        pool = Pool(self.kwargs.poolsize)
        if self.kwargs.keyfile is not None and self.kwargs.certfile is not None and self.kwargs.ca_certs is not None:
            self.__server = pywsgi.WSGIServer(
                self.getListener(),
                self.consume,
                spawn=pool,
                log=self.logger_info,
                error_log=self.logger_error,
                keyfile=self.kwargs.keyfile,
                certfile=self.kwargs.certfile,
                ca_certs=self.kwargs.ca_certs)
        else:
            self.__server = pywsgi.WSGIServer(
                self.getListener(),
                self.consume,
                spawn=pool,
                log=self.logger_info,
                error_log=self.logger_error)

        if self.kwargs.so_reuseport:
            self.__server.sock.setsockopt(socket.SOL_SOCKET, 15, 1)

        self.__server.start()
        self.logging.info("Serving on %s:%s with a connection poolsize of %s." % (self.kwargs.address, self.kwargs.port, self.kwargs.poolsize))

    def postHook(self):

        self.logging.info("Stopped serving.")
        self.__server.stop()

    def getListener(self):

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        if self.kwargs.so_reuseport:
            sock.setsockopt(socket.SOL_SOCKET, 15, 1)

        sock.setblocking(0)
        sock.bind((self.kwargs.address, self.kwargs.port))
        sock.listen(50)
        return sock




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  udpin.py
#
#  Copyright 2016 Jelle Smet development@smetj.net
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Event
from gevent.server import DatagramServer


class UDPIn(Actor):

    '''**A UDP server.**

    A UDP server.

    Parameters:

        - address(string)("0.0.0.0")
           |  The address to bind to.

        - port(int)(19283)
           |  The port to listen on.

        - reuse_port(bool)(False)
           |  Whether or not to set the SO_REUSEPORT socket option.
           |  Allows multiple instances to bind to the same port.
           |  Requires Linux kernel >= 3.9

    Queues:

        - outbox
           |  Incoming events.

    '''

    def __init__(self, actor_config, address="0.0.0.0", port=19283):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("outbox")
        self.server = DatagramServer("%s:%s" % (self.kwargs.address, self.kwargs.port), self.handle)

    def handle(self, data, address):
        '''Is called upon each incoming message'''

        event = Event(data)
        self.submit(event, self.pool.queue.outbox)

    def preHook(self):
        self.logging.info('Started listening on %s:%s' % (self.kwargs.address, self.kwargs.port))
        self.server.start()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  matchrules.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import re


class MatchRules():

    '''
    Executes different forms of matching.

    re:     Regex matching
    !re:    Negative regex matching
    >:      Bigger than
    >=:     Bigger or equal than
    <:      Smaller than
    <=:     Smaller or equal than
    =:      Equal than
    in:     Check whether element is in list
    !in:    Check whether element is not in list
    '''

    def __init__(self):
        self.methods = {"re": self.regex,
                        "!re": self.negRegex,
                        ">": self.more,
                        ">=": self.moreOrEqual,
                        "<": self.less,
                        "<=": self.lessOrEqual,
                        "=": self.equal,
                        "!=": self.notEqual,
                        "in": self.hasMember,
                        "!in": self.hasNotMember
                        }

    def do(self, condition, data):
        s = condition.split(':')
        try:
            return self.methods[s[0]](':'.join(s[1:]), data)
        except KeyError:
            raise Exception("Unknown condition")

    def regex(self, value, data):
        return bool(re.search(value, str(data)))

    def negRegex(self, value, data):
        return not bool(re.search(value, str(data)))

    def more(self, value, data):
        return float(data) > float(value)

    def moreOrEqual(self, value, data):
        return float(data) >= float(value)

    def less(self, value, data):
        return float(data) < float(value)

    def lessOrEqual(self, value, data):
        return float(data) <= float(value)

    def equal(self, value, data):
        return float(data) == float(value)

    def notEqual(self, value, data):
        return float(data) != float(value)

    def hasMember(self, value, data):
        if isinstance(data, list):
            return str(value) in data
        else:
            return False

    def hasNotMember(self, value, data):
        if isinstance(data, list):
            return str(value) not in data
        else:
            return False




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  zmqpushout.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone import Actor
import zmq.green as zmq


class ZMQPushOut(Actor):

    '''**Pushes events out to one or more ZeroMQ pull modules.**

    Expects to connect with one or more wishbone.input.push modules.  This
    module can be started in client or server mode.  In server mode, it waits
    for incoming connections.  In client mode it connects to the defined
    clients.  Events are spread in a round robin pattern over all connections.

    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - mode(str)("server")
           |  The mode to run in.  Possible options are:
           |  - server: Binds to a port and listens.
           |  - client: Connects to a port.

        - interface(string)("0.0.0.0")
           |  The interface to bind to in server <mode>.

        - port(int)(19283)
           |  The port to bind to in server <mode>.

        - clients(list)([])
           |  A list of hostname:port entries to connect to.
           |  Only valid when running in "client" <mode>.


    Queues:

        - inbox
           |  Incoming events submitted to the outside.

    '''

    def __init__(self, actor_config, selection="@data", mode="server", interface="0.0.0.0", port=19283, clients=[]):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PUSH)

        if self.kwargs.mode == "server":
            self.socket.bind("tcp://*:%s" % self.kwargs.port)
            self.logging.info("Listening on port %s" % (self.kwargs.port))
        else:
            self.socket.connect("tcp://%s" % self.kwargs.clients[0])

    def consume(self, event):

        data = event.get(self.kwargs.selection)
        try:
            # self.socket.send(event.data, flags=zmq.NOBLOCK)
            self.socket.send(data)
        except Exception as err:
            self.logging.error("Failed to submit message.  Reason: %s" % (err))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       httpinclient.py
#
#       Copyright 2016 Jelle Smet development@smetj.net
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#

from gevent import monkey; monkey.patch_socket()
import requests

from wishbone import Actor
from wishbone.event import Event
from gevent import sleep


class HTTPInClient(Actor):

    '''**A HTTP client doing http requests to pull data in.**

    This module requests an URL at the defined interval.


    Parameters:

        - url(str/list)("http://localhost")
           |  The URL to fetch (including port).
           |  When a list, will process all urls defined.

        - username(str)(None)
           |  The login to use.

        - password(str)(None)
           |  The password to use.

        - interval(int)(60)
           |  The interval in seconds between each request.

        - allow_redirects(bool(False)
           |  Allow redirects.


   Queues:

        - outbox
           |  Incoming events


    The header contains:

        - status_code:          The status code of the request.

    '''

    def __init__(self, actor_config, url="http://localhost", username=None, password=None, interval=60, allow_redirects=False):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("outbox")

    def preHook(self):
        if isinstance(self.kwargs.url, list):
            for url in self.kwargs.url:
                self.sendToBackground(self.scheduler, url)
        else:
            self.sendToBackground(self.scheduler, self.kwargs.url)

    def scheduler(self, url):
        while self.loop():

            try:
                response = requests.get(url, auth=(self.kwargs.username, self.kwargs.password), allow_redirects=self.kwargs.allow_redirects)
            except Exception as err:
                self.logging.warn("Problem requesting resource.  Reason: %s" % (err))
                sleep(1)
            else:
                event = Event(response.text)
                event.set(response.status_code, "@tmp.%s.status_code" % (self.name))
                event.set(response.url, "@tmp.%s.url" % (self.name))
                self.submit(event, self.pool.queue.outbox)
                sleep(self.kwargs.interval)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  diskin.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from gevent import sleep, event
from gevent.os import make_nonblocking
from time import time
import glob
import os
import cPickle as pickle


class DiskIn(Actor):

    '''**Reads messages from a disk buffer.**


    Parameters:

        - directory(str)
           |  The directory to write data to.

        - idle_trigger(bool)(False)
           |  When True, only reads events when no new events
           |  have been written to disk for at least <idle_time>.


        - idle_time(int)(10)
           |  The time in seconds required that no new events
           |  have been written to disk prior to start to consume
           |  the buffered data.


    Queues:

        - outbox
           |  Incoming events.
    '''

    def __init__(self, actor_config, directory="./", idle_trigger=False, idle_time=10):
        Actor.__init__(self, actor_config)
        self.reading = event.Event()
        self.reading.set()
        self.pool.createQueue("outbox")

    def preHook(self):
        self.createDir()
        self.sendToBackground(self.monitorDirectory)
        self.sendToBackground(self.diskMonitor)

    def createDir(self):

        if os.path.exists(self.kwargs.directory):
            if not os.path.isdir(self.kwargs.directory):
                raise Exception("%s exists but is not a directory" % (self.kwargs.directory))
            else:
                self.logging.info("Directory %s exists so I'm using it." % (self.kwargs.directory))
        else:
            self.logging.info("Directory %s does not exist so I'm creating it." % (self.kwargs.directory))
            os.makedirs(self.kwargs.directory)

    def monitorDirectory(self):
        while self.loop:
            self.processDirectory()
            sleep(0.5)

    def processDirectory(self):
        for filename in glob.glob("%s/*.ready" % (self.kwargs.directory)):
            self.reading.wait()
            self.readFile(filename)

    def readFile(self, filename):
        try:
            if filename.endswith("ready") and self.loop():
                with open(filename, "rb") as output_file:
                    make_nonblocking(output_file)
                    self.logging.info("Read file %s" % filename)
                    for e in self.__pickleReader(output_file):
                        self.submit(e, self.pool.queue.outbox)
                os.remove(filename)
        except Exception as err:
            self.logging.error("Failed to read file %s.  Reason: %s" % (filename, str(err)))

    def diskMonitor(self):
        '''Primitive monitor which checks whether new data is added to disk.'''

        while self.loop():
            try:
                newest = max(glob.iglob("%s/*" % (self.kwargs.directory)), key=os.path.getmtime)
            except Exception:
                pass
            else:
                if time() - os.path.getctime(newest) >= self.kwargs.idle_time:
                    self.reading.set()
                else:
                    self.reading.clear()
            sleep(1)

    def __pickleReader(self, p):

        try:
            while self.loop():
                yield pickle.load(p)
        except EOFError:
            pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  modulemanager.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import pkg_resources
import re
from prettytable import PrettyTable
from wishbone.error import ModuleInitFailure, NoSuchModule


class ModuleManager():

    def __init__(self,
                 categories=["wishbone", "wishbone.contrib"],
                 groups=["flow", "encode", "decode", "function", "input", "output"]):
        self.categories = categories
        self.groups = groups

    def exists(self, name):

        '''Returns True when module exists otherwise False'''

        if self.getModuleByName(name) is None:
            return True
        else:
            return False

    def getModule(self, category, group, name):

        m = None
        for module in pkg_resources.iter_entry_points("%s.%s" % (category, group)):
            if module.name == name:
                m = module.load()

        if m is None:
            raise NoSuchModule("Module %s.%s.%s is unknown." % (category, group, name))
        else:
            return m

    def getModulesList(self, category=None):

        if category is None:
            for category in self.categories:
                for group in self.groups:
                    group_name = "%s.%s" % (category, group)
                    groups = [m.name for m in pkg_resources.iter_entry_points(group=group_name)]
                    for m in sorted(groups):
                        yield (category, group, m)
        else:
            groups = [m.name for m in pkg_resources.iter_entry_points(group=category)]
            for m in sorted(groups):
                (c, g) = (category.split('.'))
                yield (c, g, m)

    def getModuleByName(self, name):

        (category, group, name) = name.split('.')
        return self.getModule(category, group, name)

    def getModuleDoc(self, category, group, name):

        try:
            doc = self.getModule(category, group, name).__doc__
            doc = re.search('(\*\*.*?\*\*)(.*)', doc, re.DOTALL).group(2)
            return doc
        except Exception as err:
            return "Unknown. Reason: %s" % (err)

    def getModuleTitle(self, category, group, name):

        try:
            doc = self.getModule(category, group, name).__doc__
            title = re.search('\*\*(.*?)\*\*(.*)', doc).group(1)
            return title
        except Exception as err:
            return "Unknown. Reason: %s" % (err)

    def getModuleTable(self, category=None, group=None, include_groups=[]):

        table = self.__getTable()

        category_header = None
        group_header = None
        all_items = list(self.getModulesList())

        for g in include_groups:
            all_items += list(self.getModulesList(g))

        for (category, group, module) in all_items:
            title = self.getModuleTitle(category, group, module)
            version = self.getModuleVersion(category, group, module)
            if category_header == category:
                category = ""
            else:
                category_header = category
            if group_header == group:
                group = ""
            else:
                table.add_row(["", "", "", "", ""])
                group_header = group
            table.add_row([category, group, module, version, title])

        table.add_row(["", "", "", "", ""])
        return table.get_string()

    def getModuleVersion(self, category, group, name):

        try:
            for module in pkg_resources.iter_entry_points("%s.%s" % (category, group)):
                if module.name == name:
                    return module.dist.version
        except:
            return "?"

    def validateModuleName(self, name):

        '''Validates a module reference name.'''

        if len(name.split('.')) != 3:

            raise ModuleInitFailure('%s is not a valid name structure.  Should be x.x.x' % name)

    def __getTable(self):

        t = PrettyTable(["Category", "Group", "Module", "Version", "Description"])
        t.align["Category"] = "l"
        t.align["Group"] = "l"
        t.align["Module"] = "l"
        t.align["Version"] = "r"
        t.align["Description"] = "l"

        return t



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __init__.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#
from uplook.errors import NoSuchValue

class EventLookup():
    pass


def event():
    '''Returns the requested event header value.

    - Parameters to initialize the function:

        None

    - Parameters to call the function:

        When calling the function a variable reference can be used similar to:

        ~~headerlookup("modulename.header.variablename","unknown")

        Keep in mind you always have to use a dynamic lookup function (double
        tilde).  You can provide a default value in case <variablename> does not
        exist in the header of namespace <modulename>.
    '''

    return EventLookup()


def randomword(filename=None):
    '''Returns a random word.

    This function returns a random word from a wordlist.  Each line is a new
    word.

    - Parameters to initialize the function:

        - filename(str)(None): When value is None, a buildin wordlist is used.
                               If not a filename of a wordlist is expected.


    - Parameters to call the function:

        None
    '''

    import os
    from random import choice

    class RandomWord():

        def __init__(self, f=None):
            self.wordlist = self.readWordlist()

        def readWordlist(self):
            with open("%s/../data/wordlist.txt" % (os.path.dirname(__file__))) as f:
                return f.readlines()

        def pickWord(self):
            '''Returns a word as string from the wordlist.'''

            while True:
                word = choice(self.wordlist).rstrip()
                try:
                    return word.encode("ascii", "ignore")
                except:
                    pass

    return RandomWord(filename).pickWord


def randombool():
    '''Returns True or False.

    Returns True or False randomly.

    - Parameters to initialize the function:

        None

    - Parameters to call the function:

        None
    '''

    from random import getrandbits

    def randomBool():
        return bool(getrandbits(1))

    return randomBool


def randominteger(min=0, max=100):
    '''Returns a random integer.

    Returns a random integer between <min> and <max>.

    - Parameters to initialize the function:

        - min(int)(0): The minimum value
        - max(int)(0): The maximum value

    - Parameters to call the function:

        None
    '''
    from random import randint

    def randomInteger():
        return randint(min, max)

    return randomInteger

def etcd(base="http://127.0.0.1:2379/v2/keys"):

    import requests

    base = base.rstrip('/')

    def etcdLookup(key):
        key = key.lstrip('/')

        try:
            response = requests.get('%s/%s' % (base, key))
            response.raise_for_status()
            return response.json()["node"]["value"]
        except Exception as err:
            raise NoSuchValue(str(err))
    return etcdLookup


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  zmqsubscriber.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone import Actor
from wishbone.event import Event
import zmq.green as zmq


class ZMQTopicIn(Actor):

    '''**Subscribes to one or more ZeroMQ Topic publish modules.**

    Consumes data from one or more ZeroMQ publishers.

    Parameters:

        - host(string)("localhost")
           |  The host to submit to.

        - port(int)(19283)
           |  The port to submit to.

        - timeout(int)(1)
           |  The time in seconds to timeout when connecting

        - topic(string)("")
           |  The topic to subscribe to.


    Queues:

        - outbox
           |  Incoming events.

    '''

    def __init__(self, actor_config, port=19283, timeout=10, topic=""):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("outbox")

    def preHook(self):

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.SUB)
        self.socket.connect("tcp://localhost:%s" % self.kwargs.port)
        self.socket.setsockopt(zmq.SUBSCRIBE, self.kwargs.topic)
        self.sendToBackground(self.drain)

    def drain(self):

        while self.loop():
            string = self.socket.recv()
            messagedata = string.split(" ")[1:]
            messagedata = " ".join(messagedata)
            event = Event(messagedata)
            self.submit(event, self.pool.queue.outbox)



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
#  humanlogformat.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone import Actor
from wishbone.event import Log
from time import strftime, localtime
import os
import sys


class HumanLogFormat(Actor):

    '''**Formats Wishbone log events.**

    Logs are formated from the internal wishbone format into a more
    pleasing human readable format suited for STDOUT or a logfile.

    Internal Wishbone format:

    (6, 1367682301.430527, 3342, 'Router', 'Received SIGINT. Shutting down.')

    Sample output format:

    2013-08-04T19:54:43 pid-3342 informational dictgenerator: Initiated
    2013-08-04T19:54:43 pid-3342 informational metrics_null: Started


    Parameters:

        n/a

    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, colorize=True, ident=None):
        Actor.__init__(self, actor_config)
        self.levels = {
            0: "emergency",
            1: "alert",
            2: "critical",
            3: "error",
            4: "warning",
            5: "notice",
            6: "informational",
            7: "debug"
        }
        self.colors = {
            0: "\x1B[0;35m",
            1: "\x1B[1;35m",
            2: "\x1B[0;31m",
            3: "\x1B[1;31m",
            4: "\x1B[1;33m",
            5: "\x1B[1;30m",
            6: "\x1B[1;37m",
            7: "\x1B[1;37m"
        }

        if self.kwargs.colorize:
            self.colorize = self.doColorize
        else:
            self.colorize = self.doNoColorize

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")

        if self.kwargs.ident is None:
            self.kwargs.ident = os.path.basename(sys.argv[0])

    def consume(self, event):

        data = event.get('@data')
        if isinstance(data, Log):
            log = ("%s %s %s %s: %s" % (
                strftime("%Y-%m-%dT%H:%M:%S", localtime(data.time)),
                "%s[%s]:" % (self.kwargs.ident, data.pid),
                self.levels[data.level],
                data.module,
                data.message))
            event.set(self.colorize(log, data.level))
            self.submit(event, self.pool.queue.outbox)
        else:
            raise Exception("Incoming data needs to be of type <wishbone.event.Log>. Dropped event.")

    def doColorize(self, message, level):
        return self.colors[level] + message + "\x1B[0m"

    def doNoColorize(self, message, level):
        return message



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  elasticsearchout.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.error import QueueFull
from gevent import sleep
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk


class ElasticSearchOut(Actor):

    '''**Submit data to Elasticsearch.**

    Submits data to Elasticsearch.

    Documents are indexed in bulk.  The number of messages per bulk insert is
    alligned to --queue-size.  If a number of messages sit in the buffer for
    longer than <interval> seconds, the buffer will be flushed despite not
    reaching the --queue-size number of messages.

    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - hosts(list)(["localhost:9200"])
           |  A list of "hostname:port" strings.

        - use_ssl(bool)(False)
           |  When enable expects SSL connectivity

        - verify_certs(bool)(False)
           |  When using SSL do certificate verification

        - index(str)("wishbone")
           |  The name of the index

        - doc_type(str)("wishbone")
           |  The document type

        - interval(int)(5)
           |  The buffer time flush interval.

        - flush_size(int)(100)
           |  The flush size.

    Queues:

        - inbox
           |  Incoming events submitted to the outside.

    '''

    def __init__(self, actor_config, selection="@data", hosts=["localhost:9200"], use_ssl=False, verify_certs=False, index="wishbone", doc_type="wishbone", interval=5, flush_size=100):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("inbox")
        self.pool.createQueue("bulk")
        self.pool.queue.bulk.disableFallThrough()
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):
        self.elasticsearch = Elasticsearch(self.kwargs.hosts, use_ssl=self.kwargs.use_ssl, verify_certs=self.kwargs.verify_certs)
        self.sendToBackground(self.__flushTimer)

    def consume(self, event):

        self.pool.queue.bulk.put(event)
        if self.pool.queue.bulk.size() == self.kwargs.flush_size:
            self.logging.debug("Flushing batch of %s docs after reaching batch size." % (self.kwargs.flush_size))
            self.flush()

    def flush(self):
        try:
            bulk(self.elasticsearch, [{"_index": self.kwargs.index, "_type": self.kwargs.doc_type, "_source": e.get(self.kwargs.selection)} for e in self.pool.queue.bulk.dump()])
        except Exception as err:
            self.logging.error("Failed to bulk submit messages to '%s'. Reason: %s" % (self.kwargs.hosts, err))
            for e in self.pool.queue.bulk.dump():
                self.submit(e, self.pool.queue.failed)

    def __flushTimer(self):

        while self.loop():
            sleep(self.kwargs.interval)
            if self.pool.queue.bulk.size() > 0:
                self.logging.debug("Flushing batch of %s docs after reaching timeout." % (self.pool.queue.bulk.size()))
                self.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  emailout.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone import Actor
from gevent import monkey; monkey.patch_socket()
from email.mime.text import MIMEText
import smtplib


class EmailOut(Actor):

    '''**Sends out incoming events as email.**

    Treats event.data as the body of an email.


    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - mta(string)("localhost:25)
           |  The address:port of the MTA to submit the
           |  mail to.

        - to(list)([])*
           |  A list of destinations.

        - subject(str)("Wishbone")*
           |  The subject of the email.

        - from_address(str)("wishbone@localhost")*
           |  The form email address.


    Queues:

        - inbox
           |  Incoming messages

    '''

    def __init__(self, actor_config, selection="@data", mta="localhost:25", subject="Wishbone", to=None, from_address=None):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def consume(self, event):

        try:
            message = MIMEText(str(event.get(self.kwargs.selection)))
            message["Subject"] = self.kwargs.subject
            message["From"] = self.kwargs.from_address
            message["To"] = ",".join(self.kwargs.to)

            mta = smtplib.SMTP(self.kwargs.mta)
            mta.sendmail(self.kwargs.from_address,
                         self.kwargs.to,
                         message.as_string()
                         )
        except Exception as err:
            raise Exception("Failed to send out email.  Reason: %s" % (err))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  stdout.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from os import getpid
from colorama import init, Fore, Back, Style
from gevent import monkey; monkey.patch_sys(stdin=True, stdout=True, stderr=True)
import sys


class Format():

    def __init__(self, selection, counter, pid):
        self.selection = selection
        if counter:
            self.counter = self.__returnCounter
        else:
            self.counter = self.__returnNoCounter
        if pid:
            self.pid_value = getpid()
            self.pid = self.__returnPid
        else:
            self.pid = self.__returnNoPid

    def do(self, event):
        return self.pid(self.counter(event.get(self.selection)))

    # def __returnComplete(self, event):
    #     return event.raw(complete=True)

    # def __returnIncomplete(self, event):
    #     return event.get('@data')

    def __returnCounter(self, event):
        self.countervalue += 1
        return "%s - %s" % (self.countervalue, event)

    def __returnNoCounter(self, event):
        return event

    def __returnNoPid(self, event):
        return event

    def __returnPid(self, event):
        return "PID-%s: %s" % (self.pid_value, event)


class STDOUT(Actor):

    '''**Prints incoming events to STDOUT.**

    Prints incoming events to STDOUT. When <complete> is True,
    the complete event including headers is printed to STDOUT.

    You can optionally define the colors used.


    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - counter(bool)(False)
           |  Puts an incremental number for each event in front
           |  of each event.

        - prefix(str)("")*
           |  Puts the prefix in front of each printed event.

        - pid(bool)(False)
           |  Includes the pid of the process producing the output.

        - foreground_color(str)("WHITE")
           |  The foreground color.
           |  Valid values: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE

        - background_color(str)("RESET")
           |  The background color.
           |  Valid values: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET

        - color_style(str)("NORMAL")
           |  The coloring style to use
           |  Valid values: DIM, NORMAL, BRIGHT

    Queues:

        - inbox
           |  Incoming events.
    '''

    def __init__(self, actor_config, selection="@data", counter=False, prefix="", pid=False, foreground_color="WHITE", background_color="RESET", color_style="NORMAL"):
        Actor.__init__(self, actor_config)

        self.__validateInput(foreground_color, background_color, color_style)
        self.format = Format(self.kwargs.selection, self.kwargs.counter, self.kwargs.pid)
        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

        init(autoreset=True)

    def consume(self, event):

        sys.stdout.write("%s%s%s%s%s\n" % (getattr(Fore, self.kwargs.foreground_color),
                                           getattr(Back, self.kwargs.background_color),
                                           getattr(Style, self.kwargs.color_style),
                                           self.kwargs.prefix, self.format.do(event)))
        sys.stdout.flush()

    def __validateInput(self, f, b, s):

        if f not in ["BLACK", "RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN", "WHITE"]:
            raise Exception("Foreground value is not correct.")
        if b not in ["BLACK", "RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN", "WHITE", "RESET"]:
            raise Exception("Background value is not correct.")
        if s not in ["DIM", "NORMAL", "BRIGHT"]:
            raise Exception("Style value is not correct.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  fresh.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from gevent import sleep
from wishbone.event import Event


class Fresh(Actor):

    '''**Generates a new event unless an event came through in the last x time.**

    This module just forwards events without modifying them. The moment a
    message is forwarded, the time counter is reset to its defined <timeout>
    value. If however the counter expires because no new messages have passed
    through in x time, a new message is generated with the defined <payload>.

    Parameters:

        - payload(int/float/str/obj/list/...)("wishbone")
           |  The data a generated event must contain.

        - key(str)("@data")
           |  The location to store <payload>

        - timeout(int)(60)
           |  The time in seconds Puts an incremental number for each event in front
           |  of each event.

        - payload_queue(str)("timeout")
           |  The name of the queue to submit the newly generated event to.

    Queues:

        - inbox
           |  Incoming events.

        - inbox
           |  Incoming events.

        - <payload_queue>
           |  Newly generated events due to timeout.
    '''

    def __init__(self, actor_config, payload="wishbone", key="@data", timeout=60, payload_queue="timeout"):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.pool.createQueue(self.kwargs.payload_queue)
        self.registerConsumer(self.consume, "inbox")
        self._counter = self.kwargs.timeout

    def preHook(self):

        self.sendToBackground(self.countDown)

    def consume(self, event):

        self.submit(event, self.pool.queue.outbox)
        self._resetTimeout()

    def countDown(self):

        while self.loop():
            if self._counter > 0:
                self._counter -= 1
                sleep(1)
            else:
                e = Event()
                e.set(self.kwargs.payload, self.kwargs.key)
                self.submit(e, getattr(self.pool.queue, self.kwargs.payload_queue))
                self.logging.info("Timeout of %s seconds expired.  Generting event and submit to queue %s." % (self.kwargs.timeout, self.kwargs.payload_queue))
                self._resetTimeout()

    def _resetTimeout(self):

        self._counter = self.kwargs.timeout



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  graphite.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Metric
from os.path import basename
from sys import argv
from os import getpid


class Graphite(Actor):

    '''**Converts the internal metric format to Graphite format.**

    Incoming metrics have following format:

        See http://wishbone.readthedocs.org/en/1.1.0/logs%20and%20metrics.html#format
        for more information about the format.

        Available template variables are:

            time, source, name, value unit, prefix, script, pid, source

    Parameters:

        - template(str)('{prefix}.{source}.{script}-{pid}.{type}.{name} {value} {time}'):
            | The template to use to build the metric structure.
            | Python templates are used.

        - prefix(str)("wishbone"):
            |

    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, template="{prefix}.{source}.{script}-{pid}.{type}.{name} {value} {time}", prefix='wishbone'):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")
        self.script = basename(argv[0]).replace(".py", "")
        self.pid = getpid()

    def consume(self, event):

        d = event.get()

        if isinstance(d, Metric):
            v = {"prefix": self.kwargs.prefix, "script": self.script, "pid": self.pid}
            v.update(d.dump())
            event.set(self.kwargs.template.format(**v))
            self.submit(event, self.pool.queue.outbox)
        else:
            self.logging.error("Metric dropped because not of type <wishbone.event.Metric>")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  jsonencode.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from json import dumps


class JSONEncode(Actor):

    '''**Encodes Python data objects to JSON strings.**

    Encodes Python data structures to JSON.


    Parameters:

        - source(str)("@data")
            | The data to convert.

        - destination(str)("@data")
            | The location to write the JSON string to.


    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, source='@data', destination='@data'):

        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")

    def consume(self, event):

        data = event.get(self.kwargs.source)
        data = dumps(data)
        event.set(data, self.kwargs.destination)
        self.submit(event, self.pool.queue.outbox)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  sse.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from gevent import pool
from gevent.wsgi import WSGIServer
from flask import Flask, Response, render_template_string
from gevent.queue import Queue
from uuid import uuid4
import os


class ServerSentEvent(object):

    def __init__(self, data):
        self.data = data
        self.event = None
        self.id = None
        self.desc_map = {
            self.data: "data",
            self.event: "event",
            self.id: "id"
        }

    def encode(self):
        if not self.data:
            return ""
        lines = ["%s: %s" % (v, k) for k, v in self.desc_map.iteritems() if k]
        return "%s\n\n" % "\n".join(lines)


class ServerSentEvents(Actor):

    '''**A server sent events module.**

    Sends incoming messages to a http connection.

    When the event header contains:

    event.header.<self.name>.destination then the event
    can be consumed from http://host/<destination>

    When the event header does not contain this value then
    the event can be consumed from http://host/

    If nobody is consuming events then they're lost.
    If <show_last> is set to True, then the last known value
    can be consumed by the client.

    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - bind(string)("0.0.0.0")
           |  The address to bind to.

        - port(int)(19283)
           |  The port to bind to.

        - show_last(bool)(False)
           |  If true submits the last known value to
           |  a new connection.

        - keepalive(bool)(False)
           |  If true sends a keepalive message at
           |  <keepalive_interval> seconds.

        - keepalive_interval(int)(False)
           |  The time in seconds to send keep-alive
           |  messages to the client.


    Queues:

        - inbox
           |  Incoming events submitted to the outside.

    '''

    def __init__(self, actor_config, selection="@data", bind="0.0.0.0", port=19283, show_last=False, keepalive=False, keepalive_interval=5):

        Actor.__init__(self, actor_config)
        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")
        self.session_queues = {}
        self.template = self.__getTemplate()

    def preHook(self):
        self.app = Flask(__name__)
        self.app.debug = False
        self.app.add_url_rule('/<destination>', 'target', self.target)
        self.app.add_url_rule('/', 'target', self.target)
        self.app.add_url_rule('/subscribe/<destination>', 'subscribe', self.subscribe)
        self.app.add_url_rule('/subscribe/', 'subscribe', self.subscribe)

        p = pool.Pool(1000)
        self.server = WSGIServer((self.kwargs.bind, self.kwargs.port), self.app, spawn=p, log=False)
        self.sendToBackground(self.server.serve_forever)
        self.logging.info("Listening on http://%s:%s" % (self.kwargs.bind, self.kwargs.port))

    def target(self, destination=""):
        return render_template_string(self.template, destination=destination)

    def subscribe(self, destination=""):

        def consume():
            try:
                while self.loop():
                    try:
                        result = self.session_queues[destination][queue_id].get(timeout=self.kwargs.keepalive_interval)
                    except:
                        if self.kwargs.keepalive:
                            ev = ServerSentEvent(":keep-alive")
                            yield ev.encode()
                    else:
                        ev = ServerSentEvent(str(result))
                        yield ev.encode()
            except GeneratorExit:
                self.__deleteSessionQueue(destination, queue_id)
            except Exception:
                self.__deleteSessionQueue(destination, queue_id)

        def close():
            self.__deleteSessionQueue(destination, queue_id)

        queue_id = self.__addSessionQueue(destination)
        r = Response(consume(), mimetype="text/event-stream")
        r.call_on_close(close)
        return r

    def consume(self, event):
        try:
            destination = event.get('@tmp.%s.destination' % (self.name))
        except KeyError:
            destination = ''

        try:
            for q in self.session_queues[destination]:
                self.session_queues[destination][q].put(str(event.get(self.kwargs.selection)))
        except KeyError:
            if destination == '':
                destination = '/'
            self.logging.warning("No clients are listening on %s" % destination)

    def __deleteSessionQueue(self, name, i):
        try:
            del(self.session_queues[name][i])
            if len(self.session_queues[name]) == 0:
                del(self.session_queues[name])
        except:
            pass

    def __addSessionQueue(self, name):
        if name not in self.session_queues:
            self.session_queues[name] = {}
        q = Queue()
        i = str(uuid4())
        self.session_queues[name][i] = q
        return i

    def __getTemplate(self):

        with open("%s/../data/sse.html" % (os.path.dirname(__file__))) as f:
            template = ''.join(f.readlines())
        return template




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  deserialize.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor


class Deserialize(Actor):

    '''**Creates new events from an array.**

    Takes an array and creates a new event out of each item of the array.

    Parameters:

        - source(str)("@data")
           |  The source of the array.

        - destination(str)("@data")
           |  The destination key to store the array item.

    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, source="@data", destination="@data"):

        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")

    def consume(self, event):

        data = event.get(self.kwargs.source)

        if isinstance(data, list):
            for item in data:
                e = event.clone()
                e.set(True, "@tmp.%s.generated_by" % (self.name))
                e.set("", self.kwargs.destination)
                e.set(item, self.kwargs.destination)
                self.submit(e, self.pool.queue.outbox)
        else:
            raise Exception("%s does not appear to contain an array." % (self.kwargs.source))
